#!/usr/bin/python3

import argparse
import re
import typing

parser = argparse.ArgumentParser()
parser.add_argument("filename", help="File to validate")
parser.add_argument("--verbose", help="increase output verbosity", action="store_true")
args = parser.parse_args()
filename = args.filename
verbose = args.verbose

c_and_s_regex = [("c", r"c(\d+)"), ("s", r"s(\d+)")]
short_lines_to_validate = [
    ("mr", r"_config\.mr = (\d+);"),
    ("nr", r"_config\.nr = (\d+);"),
    ("log2_kr", r"_config\.log2_kr = (\d+);"),
    ("log2_sr", r"_config\.log2_sr = (\d+);"),
]

def find_indentation(line):
    return len(line) - len(line.lstrip(' '))

def validate_code_block(f: typing.TextIO, line):
    lines_to_validate = []
    defined_variables = {}
    global line_number
    code_block_indentation = find_indentation(line)
    endgame = False
    while line:
        regex_match = re.search(r'XNN_MR_TO_INDEX\((\d+)', line)
        if regex_match and not endgame:
            first_number_should_be = regex_match.group(1)
            main_part_to_validate = line.split('__')[0].split('_')[-1] # 1x16c4
            first_number = main_part_to_validate.split('x')[0]
            second_number = re.search(r'^\d+', main_part_to_validate.split('x')[1]).group(0)
            parts_to_validate = {
                "first_number_should_be": int(first_number_should_be),
                "main_part_to_validate": main_part_to_validate,
                "first_number": int(first_number),
                "second_number": int(second_number),
            }
            # kr and sr are sometimes present
            for number_name, number_regexp in c_and_s_regex:
                regex_match = re.search(number_regexp, main_part_to_validate)
                if regex_match:
                    parts_to_validate[number_name] = int(regex_match.group(1))
            lines_to_validate.append(parts_to_validate)
        else:
            for var_name, var_regexp in short_lines_to_validate:
                regex_match = re.search(var_regexp, line)
                if regex_match:
                    endgame = True
                    defined_variables[var_name] = int(regex_match.group(1))
                    break
            else:
                if endgame:break

        if find_indentation(line) < code_block_indentation:
            # #else #endif are ok
            if line.strip() in ["#else", "#endif", ""]:
                code_block_indentation = find_indentation(line)
            else:
                break
        line = f.readline()
        line_number += 1

    # go through lines_to_validate and defined_variables and check that everything's as it should be
    first_numbers = [x["first_number"] for x in lines_to_validate]
    second_numbers = [x["second_number"] for x in lines_to_validate]
    if verbose:
        if defined_variables.get("mr") and defined_variables.get("mr") != max(first_numbers):
            yield f'mr == max(first_numbers). mr=={defined_variables.get("mr")} max(first_numbers) == {max(first_numbers)}'
    if defined_variables.get("nr") and {defined_variables.get("nr")} != set(second_numbers):
        yield f'nr should be same as all second_numbers. nr=={defined_variables.get("nr")} set(second_numbers) == {set(second_numbers)}'
    c_value = lines_to_validate[0].get("c")
    s_value = lines_to_validate[0].get("s")
    for line_to_validate in lines_to_validate:
        if line_to_validate.get("first_number_should_be") != line_to_validate.get("first_number"):
            yield (f"XNN_MR_TO_INDEX(x) is not the same as the first number."
                  f" they are {line_to_validate.get('first_number_should_be')} and {line_to_validate.get('first_number')}")
        # if c or r are present for one they should be present for all
        if line_to_validate.get('c') != c_value:
            yield f"c is not the same as previous lines. previously was {c_value}. now is {line_to_validate.get('c')}"
        if line_to_validate.get("s") != s_value:
            yield f"s is not the same as previous lines. previously was {s_value}. now is {line_to_validate.get('s')}"

    # check that sr and kr are there when they should be there (and not there when they shouldn't be)
    if (not (defined_variables.get("log2_kr") == c_value == None) and
            (defined_variables.get("log2_kr") is None or 2 ** defined_variables.get("log2_kr") != c_value)):
        if (defined_variables.get("log2_kr") != None and c_value != None) or verbose:
            yield f"log2_kr does not match c. log2_kr == {defined_variables.get('log2_kr')}. c == {c_value}"
    if (not (defined_variables.get("log2_sr") == s_value == None) and
            (defined_variables.get("log2_sr") is None or 2 ** defined_variables.get("log2_sr") != s_value)):
        if (defined_variables.get("log2_sr") != None and s_value != None) or verbose:
            yield f"log2_sr does not match r. log2_sr == {defined_variables.get('log2_sr')}. s == {s_value}"


with open(filename, 'r') as f:
    line = f.readline()
    line_number = 1
    while line:
        if "XNN_MR_TO_INDEX" in line and line.startswith("      "):
            for warning in validate_code_block(f, line):
                print(f'in file {filename} near line {line_number}: {warning}')
        line = f.readline()
        line_number += 1



#!/usr/bin/env python
# Copyright 2019 Google LLC
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import codecs
import os


def _indent(text):
  return "\n".join(map(lambda t: "  " + t if t else t, text.splitlines()))


def _remove_duplicate_newlines(text):
  filtered_lines = list()
  last_newline = False
  for line in text.splitlines():
    is_newline = len(line.strip()) == 0
    if not is_newline or not last_newline:
      filtered_lines.append(line)
    last_newline = is_newline
  return "\n".join(filtered_lines)

_XNNPACK_SRC = ""

_ARCH_TO_MACRO_MAP = {
  "aarch32": "XNN_ARCH_ARM",
  "aarch64": "XNN_ARCH_ARM64",
  "x86-32": "XNN_ARCH_X86",
  "x86-64": "XNN_ARCH_X86_64",
  "hexagon": "XNN_ARCH_HEXAGON",
  "riscv": "XNN_ARCH_RISCV",
  "wasm": "XNN_ARCH_WASM",
  "wasmsimd": "XNN_ARCH_WASMSIMD",
  "wasmrelaxedsimd": "XNN_ARCH_WASMRELAXEDSIMD",
  "wasm32": "XNN_ARCH_WASM",
  "wasmsimd32": "XNN_ARCH_WASMSIMD",
  "wasmrelaxedsimd32": "XNN_ARCH_WASMRELAXEDSIMD",
}

# Mapping from ISA extension to macro guarding build-time enabled/disabled
# status for the ISA. Only ISAs that can be enabled/disabled have an entry.
_ISA_TO_MACRO_MAP = {
  "fp16arith": "XNN_ENABLE_ARM_FP16_SCALAR",
  "neonfp16arith": "XNN_ENABLE_ARM_FP16_VECTOR",
  "neonbf16": "XNN_ENABLE_ARM_BF16",
  "neondot": "XNN_ENABLE_ARM_DOTPROD",
  "neondotfp16arith": "XNN_ENABLE_ARM_DOTPROD && XNN_ENABLE_ARM_FP16_VECTOR",
  "neoni8mm": "XNN_ENABLE_ARM_I8MM",
  "rvv": "XNN_ENABLE_RISCV_VECTOR",
  "rvvfp16arith": "XNN_ENABLE_RISCV_FP16_VECTOR",
  "avxvnni": "XNN_ENABLE_AVXVNNI",
  "avxvnniint8": "XNN_ENABLE_AVXVNNIINT8",
  "avx256skx": "XNN_ENABLE_AVX256SKX",
  "avx256vnni": "XNN_ENABLE_AVX256VNNI",
  "avx256vnnigfni": "XNN_ENABLE_AVX256VNNIGFNI",
  "avx512f": "XNN_ENABLE_AVX512F",
  "avx512skx": "XNN_ENABLE_AVX512SKX",
  "avx512vbmi": "XNN_ENABLE_AVX512VBMI",
  "avx512vnni": "XNN_ENABLE_AVX512VNNI",
  "avx512vnnigfni": "XNN_ENABLE_AVX512VNNIGFNI",
  "avx512amx": "XNN_ENABLE_AVX512AMX",
  "avx512fp16": "XNN_ENABLE_AVX512FP16",
  "hvx": "XNN_ENABLE_HVX",
}

_ISA_TO_ARCH_MAP = {
  "armsimd32": ["aarch32"],
  "fp16arith": ["aarch32", "aarch64"],
  "neon": ["aarch32", "aarch64"],
  "neonfp16": ["aarch32", "aarch64"],
  "neonfma": ["aarch32", "aarch64"],
  "neonv8": ["aarch32", "aarch64"],
  "neonfp16arith": ["aarch32", "aarch64"],
  "neonbf16": ["aarch32", "aarch64"],
  "neondot": ["aarch32", "aarch64"],
  "neondotfp16arith": ["aarch32", "aarch64"],
  "neoni8mm": ["aarch64"],
  "sse": ["x86-32", "x86-64"],
  "sse2": ["x86-32", "x86-64"],
  "ssse3": ["x86-32", "x86-64"],
  "sse41": ["x86-32", "x86-64"],
  "avx": ["x86-32", "x86-64"],
  "f16c": ["x86-32", "x86-64"],
  "fma3": ["x86-32", "x86-64"],
  "avx2": ["x86-32", "x86-64"],
  "avx512f": ["x86-32", "x86-64"],
  "avx512skx": ["x86-32", "x86-64"],
  "avx512vbmi": ["x86-32", "x86-64"],
  "avx512vnni": ["x86-32", "x86-64"],
  "avx512vnnigfni": ["x86-32", "x86-64"],
  "avx512amx": ["x86-32", "x86-64"],
  "avx512fp16": ["x86-32", "x86-64"],
  "avxvnni": ["x86-32", "x86-64"],
  "avxvnniint8": ["x86-32", "x86-64"],
  "avx256skx": ["x86-32", "x86-64"],
  "avx256vnni": ["x86-32", "x86-64"],
  "avx256vnnigfni": ["x86-32", "x86-64"],
  "hexagon": ["hexagon"],
  "hvx": ["hexagon"],
  "rvv": ["riscv"],
  "rvvfp16arith": ["riscv"],
  "wasm32": ["wasm", "wasmsimd"],
  "wasm": ["wasm", "wasmsimd", "wasmrelaxedsimd"],
  "wasmsimd": ["wasmsimd", "wasmrelaxedsimd"],
  "wasmrelaxedsimd": ["wasmrelaxedsimd"],
  "wasmpshufb": ["wasmrelaxedsimd"],
  "wasmsdot": ["wasmrelaxedsimd"],
  "wasmusdot": ["wasmrelaxedsimd"],
  "wasmblendvps": ["wasmrelaxedsimd"],
}

_ISA_TO_UTILCHECK_MAP = {
  "armsimd32": "CheckARMV6",
  "fp16arith": "CheckFP16ARITH",
  "neon": "CheckNEON",
  "neonfp16": "CheckNEONFP16",
  "neonfma": "CheckNEONFMA",
  "neonv8": "CheckNEONV8",
  "neonfp16arith": "CheckNEONFP16ARITH",
  "neonbf16": "CheckNEONBF16",
  "neondot": "CheckNEONDOT",
  "neondotfp16arith": "CheckNEONDOT",
  "neoni8mm": "CheckNEONI8MM",
  "ssse3": "CheckSSSE3",
  "sse41": "CheckSSE41",
  "avx": "CheckAVX",
  "f16c": "CheckF16C",
  "avx2": "CheckAVX2",
  "fma3": "CheckFMA3",
  "avx512f": "CheckAVX512F",
  "avx512skx": "CheckAVX512SKX",
  "avx512vbmi": "CheckAVX512VBMI",
  "avx512vnni": "CheckAVX512VNNI",
  "avx512vnnigfni": "CheckAVX512VNNIGFNI",
  "avx512amx": "CheckAVX512AMX",
  "avx512fp16": "CheckAVX512FP16",
  "avxvnni": "CheckAVXVNNI",
  "avxvnniint8": "CheckAVXVNNIINT8",
  "avx256skx": "CheckAVX256SKX",
  "avx256vnni": "CheckAVX256VNNI",
  "avx256vnnigfni": "CheckAVX256VNNIGFNI",
  "hvx": "CheckHVX",
  "rvv": "CheckRVV",
  "rvvfp16arith": "CheckRVVFP16ARITH",
  "wasmpshufb": "CheckWAsmPSHUFB",
  "wasmsdot": "CheckWAsmSDOT",
  "wasmusdot": "CheckWAsmUSDOT",
  "wasmblendvps": "CheckWAsmBLENDVPS",
}

_ISA_TO_CHECK_MAP = {
  "armsimd32": "TEST_REQUIRES_ARM_SIMD32",
  "fp16arith": "TEST_REQUIRES_ARM_FP16_ARITH",
  "neon": "TEST_REQUIRES_ARM_NEON",
  "neonfp16": "TEST_REQUIRES_ARM_NEON_FP16",
  "neonfma": "TEST_REQUIRES_ARM_NEON_FMA",
  "neonv8": "TEST_REQUIRES_ARM_NEON_V8",
  "neonfp16arith": "TEST_REQUIRES_ARM_NEON_FP16_ARITH",
  "neonbf16": "TEST_REQUIRES_ARM_NEON_BF16",
  "neondot": "TEST_REQUIRES_ARM_NEON_DOT",
  "neondotfp16arith": "TEST_REQUIRES_ARM_NEON_DOT_FP16_ARITH",
  "neoni8mm": "TEST_REQUIRES_ARM_NEON_I8MM",
  "sse": "TEST_REQUIRES_X86_SSE",
  "sse2": "TEST_REQUIRES_X86_SSE2",
  "ssse3": "TEST_REQUIRES_X86_SSSE3",
  "sse41": "TEST_REQUIRES_X86_SSE41",
  "avx": "TEST_REQUIRES_X86_AVX",
  "f16c": "TEST_REQUIRES_X86_F16C",
  "avx2": "TEST_REQUIRES_X86_AVX2",
  "fma3": "TEST_REQUIRES_X86_FMA3",
  "avx512f": "TEST_REQUIRES_X86_AVX512F",
  "avx512skx": "TEST_REQUIRES_X86_AVX512SKX",
  "avx512vbmi": "TEST_REQUIRES_X86_AVX512VBMI",
  "avx512vnni": "TEST_REQUIRES_X86_AVX512VNNI",
  "avx512vnnigfni": "TEST_REQUIRES_X86_AVX512VNNIGFNI",
  "avx512amx": "TEST_REQUIRES_X86_AVX512AMX",
  "avx512fp16": "TEST_REQUIRES_X86_AVX512FP16",
  "avxvnni": "TEST_REQUIRES_X86_AVXVNNI",
  "avxvnniint8": "TEST_REQUIRES_X86_AVXVNNIINT8",
  "avx256skx": "TEST_REQUIRES_X86_AVX256SKX",
  "avx256vnni": "TEST_REQUIRES_X86_AVX256VNNI",
  "avx256vnnigfni": "TEST_REQUIRES_X86_AVX256VNNIGFNI",
  "hvx": "TEST_REQUIRES_HVX",
  "rvv": "TEST_REQUIRES_RISCV_VECTOR",
  "rvvfp16arith": "TEST_REQUIRES_RISCV_VECTOR_FP16_ARITH",
  "wasmpshufb": "TEST_REQUIRES_WASM_PSHUFB",
  "wasmsdot": "TEST_REQUIRES_WASM_SDOT",
  "wasmusdot": "TEST_REQUIRES_WASM_USDOT",
  "wasmblendvps": "TEST_REQUIRES_WASM_BLENDVPS",
}


def xnnpack_src():
  return _XNNPACK_SRC


def isa_hierarchy_map():
  return _ISA_HIERARCHY_MAP


def parse_target_name(target_name):
  arch = list()
  isa = None
  assembly = False
  for target_part in target_name.split("_"):
    if target_part in _ARCH_TO_MACRO_MAP:
      if target_part in _ISA_TO_ARCH_MAP:
        arch = _ISA_TO_ARCH_MAP[target_part]
        isa = target_part
      else:
        arch = [target_part]
    elif target_part in _ISA_TO_ARCH_MAP:
      isa = target_part
    elif target_part == "asm":
      assembly = True
  if isa and not arch:
    arch = _ISA_TO_ARCH_MAP[isa]

  return arch, isa, assembly


def generate_isa_check_macro(isa):
  return _ISA_TO_CHECK_MAP.get(isa, "")

def generate_isa_utilcheck_macro(isa):
  return _ISA_TO_UTILCHECK_MAP.get(isa, "")

def arch_to_macro(arch, isa):
  return _ARCH_TO_MACRO_MAP[arch]

def postprocess_test_case(test_case, arch, isa, assembly=False):
  test_case = _remove_duplicate_newlines(test_case)
  if arch:
    guard = " || ".join(arch_to_macro(a, isa) for a in arch)
    if isa in _ISA_TO_MACRO_MAP:
      if len(arch) > 1:
        guard = "%s && (%s)" % (_ISA_TO_MACRO_MAP[isa], guard)
      else:
        guard = "%s && %s" % (_ISA_TO_MACRO_MAP[isa], guard)
    if assembly and "||" in guard:
      guard = '(' + guard + ')'
    if assembly:
      guard += " && XNN_ENABLE_ASSEMBLY"
    return "#if %s\n" % guard + _indent(test_case) + "\n" + \
      "#endif  // %s\n" % guard
  else:
    return test_case

_ISA_HIERARCHY = [
  "sse",
  "sse2",
  "ssse3",
  "sse41",
  "avx",
  "avx2",
  "avx512f",
  "avx512vbmi",
  "avxvnni",
  "avxvnniint8",
  "avx256skx",
  "avx256vnni",
  "avx256vnnigfni",
  "avx512skx",
  "avx512vnni",
  "avx512vnnigfni",
  "avx512fp16",
  "avx512amx",
  "armsimd32",
  "neon",
  "neonv8",
  "neondot",
  "neondotfp16",
  "neoni8mm",
  "wasm",
  "wasmsimd",
  "wasmrelaxedsimd",
  "rvv",
  "rvvfp16",
  "hexagon",
  "hvx"
]

_ISA_HIERARCHY_MAP = {isa: v for v, isa in enumerate(_ISA_HIERARCHY)}


def overwrite_if_changed(filepath, content):
  txt_changed = True
  if os.path.exists(filepath):
    with codecs.open(filepath, "r", encoding="utf-8") as output_file:
      txt_changed = output_file.read() != content
  if txt_changed:
    with codecs.open(filepath, "w", encoding="utf-8") as output_file:
      output_file.write(content)

def make_multiline_macro(x):
  lines = x.strip().split('\n')
  max_len = max([len(i) for i in lines])
  lines = [i.ljust(max_len) + "\\" for i in lines]
  return "\n".join(lines)[:-1].strip() + "\n"


#!/usr/bin/env python
# Copyright 2023 Google LLC
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import codecs
import os
import re
import sys
import yaml

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
from primes import next_prime
import xngen
import xnncommon


parser = argparse.ArgumentParser(
    description="Test generator for CONV HWC2CHW micro-kernels"
)
parser.add_argument(
    "-s", "--spec", metavar="FILE", required=True, help="Spec (YAML) file"
)
parser.add_argument(
    "-o",
    "--output",
    metavar="FILE",
    required=True,
    help="Output (C++ source) file",
)
parser.set_defaults(defines=list())


TEST_TEMPLATE = """\
TEST(${TEST_NAME}, input_width_eq_${INPUT_WIDTH}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  ConvHWC2CHWMicrokernelTester()
    .kernel_size(${KERNEL_SIZE})
    .subsampling(${SUBSAMPLING})
    .padding_width(${PADDING_RIGHT})
    .input_channels(${INPUT_CHANNELS})
    .output_channels_tile(${OUTPUT_CHANNELS_TILE})
    .output_channels(${OUTPUT_CHANNELS_TILE})
    .input_width(${INPUT_WIDTH})
    .input_height(${KERNEL_SIZE})
    .Test(${", ".join(TEST_ARGS)});
}

$if INPUT_WIDTH > 1:
  TEST(${TEST_NAME}, input_width_div_${INPUT_WIDTH}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t input_width = ${INPUT_WIDTH*2}; input_width <= ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*3}) {
      ConvHWC2CHWMicrokernelTester()
        .kernel_size(${KERNEL_SIZE})
        .subsampling(${SUBSAMPLING})
        .padding_width(${PADDING_RIGHT})
        .input_channels(${INPUT_CHANNELS})
        .output_channels_tile(${OUTPUT_CHANNELS_TILE})
        .output_channels(${OUTPUT_CHANNELS_TILE})
        .input_width(input_width)
        .input_height(${KERNEL_SIZE})
        .Test(${", ".join(TEST_ARGS)});
    }
  }

  TEST(${TEST_NAME}, input_width_lt_${INPUT_WIDTH}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH}; input_width++) {
      ConvHWC2CHWMicrokernelTester()
        .kernel_size(${KERNEL_SIZE})
        .subsampling(${SUBSAMPLING})
        .padding_width(${PADDING_RIGHT})
        .input_channels(${INPUT_CHANNELS})
        .output_channels_tile(${OUTPUT_CHANNELS_TILE})
        .output_channels(${OUTPUT_CHANNELS_TILE})
        .input_width(input_width)
        .input_height(${KERNEL_SIZE})
        .Test(${", ".join(TEST_ARGS)});
    }
  }

TEST(${TEST_NAME}, input_width_gt_${INPUT_WIDTH}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t input_width = ${INPUT_WIDTH+1}; input_width < ${33 if INPUT_WIDTH == 1 else INPUT_WIDTH*2}; input_width++) {
    ConvHWC2CHWMicrokernelTester()
      .kernel_size(${KERNEL_SIZE})
      .subsampling(${SUBSAMPLING})
      .padding_width(${PADDING_RIGHT})
      .input_channels(${INPUT_CHANNELS})
      .output_channels_tile(${OUTPUT_CHANNELS_TILE})
      .output_channels(${OUTPUT_CHANNELS_TILE})
      .input_width(input_width)
      .input_height(${KERNEL_SIZE})
      .Test(${", ".join(TEST_ARGS)});
  }
}

TEST(${TEST_NAME}, output_channels_lt_${OUTPUT_CHANNELS_TILE}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t output_channels = 1; output_channels < ${OUTPUT_CHANNELS_TILE}; output_channels++) {
    for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
      ConvHWC2CHWMicrokernelTester()
        .kernel_size(${KERNEL_SIZE})
        .subsampling(${SUBSAMPLING})
        .padding_width(${PADDING_RIGHT})
        .input_channels(${INPUT_CHANNELS})
        .output_channels_tile(${OUTPUT_CHANNELS_TILE})
        .output_channels(output_channels)
        .input_width(input_width)
        .input_height(${KERNEL_SIZE})
        .Test(${", ".join(TEST_ARGS)});
    }
  }
}

TEST(${TEST_NAME}, output_channels_div_${OUTPUT_CHANNELS_TILE}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t output_channels = ${OUTPUT_CHANNELS_TILE*2}; output_channels <= ${OUTPUT_CHANNELS_TILE*4}; output_channels += ${OUTPUT_CHANNELS_TILE}) {
    for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
      ConvHWC2CHWMicrokernelTester()
        .kernel_size(${KERNEL_SIZE})
        .subsampling(${SUBSAMPLING})
        .padding_width(${PADDING_RIGHT})
        .input_channels(${INPUT_CHANNELS})
        .output_channels_tile(${OUTPUT_CHANNELS_TILE})
        .output_channels(output_channels)
        .input_width(input_width)
        .input_height(${KERNEL_SIZE})
        .Test(${", ".join(TEST_ARGS)});
    }
  }
}

TEST(${TEST_NAME}, output_channels_gt_${OUTPUT_CHANNELS_TILE}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t output_channels = ${OUTPUT_CHANNELS_TILE+1}; output_channels < ${OUTPUT_CHANNELS_TILE*2}; output_channels++) {
    for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
      ConvHWC2CHWMicrokernelTester()
        .kernel_size(${KERNEL_SIZE})
        .subsampling(${SUBSAMPLING})
        .padding_width(${PADDING_RIGHT})
        .input_channels(${INPUT_CHANNELS})
        .output_channels_tile(${OUTPUT_CHANNELS_TILE})
        .output_channels(output_channels)
        .input_width(input_width)
        .input_height(${KERNEL_SIZE})
        .Test(${", ".join(TEST_ARGS)});
    }
  }
}

TEST(${TEST_NAME}, input_height_lt_3) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t input_height = 1; input_height < 3; input_height++) {
    for (size_t output_channels = 1; output_channels < ${OUTPUT_CHANNELS_TILE*2}; output_channels += ${OUTPUT_CHANNELS_TILE-1}) {
      for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
        ConvHWC2CHWMicrokernelTester()
          .kernel_size(${KERNEL_SIZE})
          .subsampling(${SUBSAMPLING})
          .padding(1)  // padded input height of at least 3 required
          .input_channels(${INPUT_CHANNELS})
          .output_channels_tile(${OUTPUT_CHANNELS_TILE})
          .output_channels(output_channels)
          .input_width(input_width)
          .input_height(input_height)
          .Test(${", ".join(TEST_ARGS)});
      }
    }
  }
}

TEST(${TEST_NAME}, input_height_gt_3) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t input_height = 4; input_height <= 9; input_height++) {
    for (size_t output_channels = 1; output_channels < ${OUTPUT_CHANNELS_TILE*2}; output_channels += ${OUTPUT_CHANNELS_TILE-1}) {
      for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
        ConvHWC2CHWMicrokernelTester()
          .kernel_size(${KERNEL_SIZE})
          .subsampling(${SUBSAMPLING})
          .padding_width(${PADDING_RIGHT})
          .input_channels(${INPUT_CHANNELS})
          .output_channels_tile(${OUTPUT_CHANNELS_TILE})
          .output_channels(output_channels)
          .input_width(input_width)
          .input_height(input_height)
          .Test(${", ".join(TEST_ARGS)});
      }
    }
  }
}

TEST(${TEST_NAME}, padding_top) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t padding_top = 0; padding_top <= 1; padding_top++) {
    for (size_t output_channels = 1; output_channels < ${OUTPUT_CHANNELS_TILE*4}; output_channels += ${OUTPUT_CHANNELS_TILE*2-1}) {
      for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
        ConvHWC2CHWMicrokernelTester()
          .kernel_size(${KERNEL_SIZE})
          .subsampling(${SUBSAMPLING})
          .padding_width(${PADDING_RIGHT})
          .padding_top(padding_top)
          .input_channels(${INPUT_CHANNELS})
          .output_channels_tile(${OUTPUT_CHANNELS_TILE})
          .output_channels(output_channels)
          .input_width(input_width)
          .input_height(9)
          .Test(${", ".join(TEST_ARGS)});
      }
    }
  }
}

TEST(${TEST_NAME}, padding_bottom) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t padding_bottom = 0; padding_bottom <= 1; padding_bottom++) {
    for (size_t output_channels = 1; output_channels < ${OUTPUT_CHANNELS_TILE*4}; output_channels += ${OUTPUT_CHANNELS_TILE*2-1}) {
      for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
        ConvHWC2CHWMicrokernelTester()
          .kernel_size(${KERNEL_SIZE})
          .subsampling(${SUBSAMPLING})
          .padding_width(${PADDING_RIGHT})
          .padding_bottom(padding_bottom)
          .input_channels(${INPUT_CHANNELS})
          .output_channels_tile(${OUTPUT_CHANNELS_TILE})
          .output_channels(output_channels)
          .input_width(input_width)
          .input_height(9)
          .Test(${", ".join(TEST_ARGS)});
      }
    }
  }
}

TEST(${TEST_NAME}, output_y_start) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t output_y_start = 1; output_y_start <= 3; output_y_start++) {
    for (size_t output_channels = 1; output_channels < ${OUTPUT_CHANNELS_TILE*2}; output_channels += ${OUTPUT_CHANNELS_TILE-1}) {
      for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
        ConvHWC2CHWMicrokernelTester()
          .kernel_size(${KERNEL_SIZE})
          .subsampling(${SUBSAMPLING})
          .padding_width(${PADDING_RIGHT})
          .input_channels(${INPUT_CHANNELS})
          .output_channels_tile(${OUTPUT_CHANNELS_TILE})
          .output_channels(output_channels)
          .input_width(input_width)
          .input_height(9)
          .output_y_start(output_y_start)
          .Test(${", ".join(TEST_ARGS)});
      }
    }
  }
}

TEST(${TEST_NAME}, output_y_end) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t output_y_end = 2; output_y_end < 5; output_y_end++) {
    for (size_t output_channels = 1; output_channels < ${OUTPUT_CHANNELS_TILE*2}; output_channels += ${OUTPUT_CHANNELS_TILE-1}) {
      for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
        ConvHWC2CHWMicrokernelTester()
          .kernel_size(${KERNEL_SIZE})
          .subsampling(${SUBSAMPLING})
          .padding_width(${PADDING_RIGHT})
          .input_channels(${INPUT_CHANNELS})
          .output_channels_tile(${OUTPUT_CHANNELS_TILE})
          .output_channels(output_channels)
          .input_width(input_width)
          .input_height(9)
          .output_y_end(output_y_end)
          .Test(${", ".join(TEST_ARGS)});
      }
    }
  }
}

TEST(${TEST_NAME}, qmin) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t output_channels = 1; output_channels < ${OUTPUT_CHANNELS_TILE*2}; output_channels += ${OUTPUT_CHANNELS_TILE-1}) {
    for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
      ConvHWC2CHWMicrokernelTester()
        .kernel_size(${KERNEL_SIZE})
        .subsampling(${SUBSAMPLING})
        .padding_width(${PADDING_RIGHT})
        .input_channels(${INPUT_CHANNELS})
        .output_channels_tile(${OUTPUT_CHANNELS_TILE})
        .output_channels(output_channels)
        .input_width(input_width)
        .input_height(6)
        .qmin(128)
        .Test(${", ".join(TEST_ARGS)});
    }
  }
}

TEST(${TEST_NAME}, qmax) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t output_channels = 1; output_channels < ${OUTPUT_CHANNELS_TILE*2}; output_channels += ${OUTPUT_CHANNELS_TILE-1}) {
    for (size_t input_width = ${1 if PADDING_LEFT else 2}; input_width < ${INPUT_WIDTH*8}; input_width += ${INPUT_WIDTH*2-1}) {
      ConvHWC2CHWMicrokernelTester()
        .kernel_size(${KERNEL_SIZE})
        .subsampling(${SUBSAMPLING})
        .padding_width(${PADDING_RIGHT})
        .input_channels(${INPUT_CHANNELS})
        .output_channels_tile(${OUTPUT_CHANNELS_TILE})
        .output_channels(output_channels)
        .input_width(input_width)
        .input_height(6)
        .qmax(128)
        .Test(${", ".join(TEST_ARGS)});
    }
  }
}
"""


def split_ukernel_name(name):
  match = re.fullmatch(
      r"xnn_(f16|f32)_conv_hwc2chw_ukernel_(\d+)x(\d+)s(\d+)(p1)c(\d+)x(\d+)__(.+)_(\d+)x(\d+)?",
      name,
  )
  assert match is not None
  kernel_height, kernel_width = int(match.group(2)), int(match.group(3))
  assert kernel_height == kernel_width
  subsampling = int(match.group(4))
  padding_right = 1
  padding_left = 1
  input_channels = int(match.group(6))
  channel_tile = int(match.group(7))
  height_tile = int(match.group(9))
  width_tile = int(match.group(10))

  arch, isa, assembly = xnncommon.parse_target_name(target_name=match.group(8))
  return (
      kernel_height,
      kernel_width,
      subsampling,
      padding_left,
      padding_right,
      input_channels,
      channel_tile,
      height_tile,
      width_tile,
      arch,
      isa,
  )


def generate_test_cases(
    ukernel,
    kernel_height,
    kernel_width,
    subsampling,
    padding_left,
    padding_right,
    input_channels,
    channel_tile,
    height_tile,
    width_tile,
    init_fn,
    isa,
):
  """Generates all tests cases for a CONV HWC2CHW micro-kernel.

  Args:
    ukernel: C name of the micro-kernel function.
    kernel_height: convolution kernel height assumed by the micro-kernel.
    kernel_width: convolution kernel width assumed by the micro-kernel.
    subsampling: convolution subsampling (stride) assumed by the micro-kernel.
      The same subsampling factor is assumed for both horizontal and vertical
      directions.
    padding_left: input padding on the left assumed by micro-kernel, 0 or 1
    padding_right: input padding on the right assumed by micro-kernel, current
      microkernels always assume this is 1
    input_channels: number of input channels assumed by micro-kernel.
    channel_tile: number of output channels processed in one iteration of the
      main loop of the micro-kernel.
    height_tile: number of output rows processed in one iteration of the main
      loop of the micro-kernel.
    width_tile: number of output columns processed in one iteration of the main
      loop of the micro-kernel.
    init_fn: C name of the function to initialize microkernel parameters.
    isa: instruction set required to run the micro-kernel. Generated unit test
      will skip execution if the host processor doesn't support this ISA.

  Returns:
    Code for the test case.
  """
  _, test_name = ukernel.split("_", 1)
  _, datatype, ukernel_type, _ = ukernel.split("_", 3)
  test_args = [ukernel, init_fn]
  return xngen.preprocess(
      TEST_TEMPLATE,
      {
          "TEST_NAME": test_name.upper().replace("UKERNEL_", ""),
          "TEST_ARGS": test_args,
          "UKERNEL_TYPE": ukernel_type.upper(),
          "DATATYPE": datatype,
          "INPUT_WIDTH": width_tile * height_tile,
          "KERNEL_SIZE": kernel_height,
          "KERNEL_HEIGHT": kernel_height,
          "KERNEL_WIDTH": kernel_width,
          "SUBSAMPLING": subsampling,
          "PADDING_LEFT": padding_left,
          "PADDING_RIGHT": padding_right,
          "INPUT_CHANNELS": input_channels,
          "OUTPUT_CHANNELS_TILE": channel_tile,
          "HEIGHT_TILE": height_tile,
          "WIDTH_TILE": width_tile,
          "ISA_CHECK": xnncommon.generate_isa_check_macro(isa),
          "next_prime": next_prime,
      },
  )


def main(args):
  options = parser.parse_args(args)

  with codecs.open(options.spec, "r", encoding="utf-8") as spec_file:
    spec_yaml = yaml.safe_load(spec_file)
    if not isinstance(spec_yaml, list):
      raise ValueError("expected a list of micro-kernels in the spec")

    tests = """\
// Copyright 2023 Google LLC
//
// This source code is licensed under the BSD-style license found in the
// LICENSE file in the root directory of this source tree.
//
// Auto-generated file. Do not edit!
//   Specification: {specification}
//   Generator: {generator}


#include <gtest/gtest.h>
#include "xnnpack/common.h"
#include "xnnpack/conv.h"
#include "xnnpack/isa-checks.h"
#include "xnnpack/microparams-init.h"
#include "conv-hwc2chw-microkernel-tester.h"
""".format(specification=options.spec, generator=sys.argv[0])

    for ukernel_spec in spec_yaml:
      name = ukernel_spec["name"]
      init_fn = ukernel_spec["init"]
      (
          kernel_height,
          kernel_width,
          subsampling,
          padding_left,
          padding_right,
          input_channels,
          channel_tile,
          height_tile,
          width_tile,
          arch,
          isa,
      ) = split_ukernel_name(name)

      test_case = generate_test_cases(
          name,
          kernel_height,
          kernel_width,
          subsampling,
          padding_left,
          padding_right,
          input_channels,
          channel_tile,
          height_tile,
          width_tile,
          init_fn,
          isa,
      )
      tests += "\n\n" + xnncommon.postprocess_test_case(test_case, arch, isa)

    xnncommon.overwrite_if_changed(options.output, tests)


if __name__ == "__main__":
  main(sys.argv[1:])



#!/usr/bin/env python
# Copyright 2019 Google LLC
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import codecs
import math
import os
import re
import sys
import yaml

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
import xngen
import xnncommon


parser = argparse.ArgumentParser(
  description='Vector binary operation microkernel test generator')
parser.add_argument("-t", "--tester", metavar="TESTER", required=True,
                    choices=[
                    "VCMulMicrokernelTester",
                    "VBinaryMicrokernelTester"],
                    help="Tester class to be used in the generated test")
parser.add_argument("-b", "--broadcast_b", action="store_true",
                    help='Broadcast the RHS of the operation')
parser.add_argument("-k", "--ukernel", required=True,
                    help="Microkernel type")
parser.add_argument("-o", "--output", metavar="FILE", required=True,
                    help='Output (C++ source) file')
parser.set_defaults(defines=list())

OP_TYPES = {
    "vadd": "Add",
    "vaddc": "Add",
    "vcopysign": "CopySign",
    "vcopysignc": "CopySign",
    "vrcopysign": "RCopySign",
    "vrcopysignc": "RCopySign",
    "vdiv": "Div",
    "vdivc": "Div",
    "vrdiv": "RDiv",
    "vrdivc": "RDiv",
    "vmax": "Max",
    "vmaxc": "Max",
    "vmin": "Min",
    "vminc": "Min",
    "vmul": "Mul",
    "vmulc": "Mul",
    "vcmul": "CMul",
    "vsub": "Sub",
    "vsubc": "Sub",
    "vrsub": "RSub",
    "vrsubc": "RSub",
    "vsqrdiff": "SqrDiff",
    "vsqrdiffc": "SqrDiff",
    "vprelu": "Prelu",
    "vpreluc": "Prelu",
    "vrpreluc": "RPrelu",
}

BINOP_TEST_TEMPLATE = """
#define XNN_UKERNEL_WITH_PARAMS(arch_flags, ukernel, batch_tile, vector_tile, datatype, params_type, init_params)
XNN_TEST_BINARY_BATCH_EQ(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
XNN_TEST_BINARY_BATCH_DIV(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
XNN_TEST_BINARY_BATCH_LT(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
XNN_TEST_BINARY_BATCH_GT(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});

$if TESTER in ["VMulCMicrokernelTester"]:
  XNN_TEST_BINARY_INPLACE_A(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
$elif ${BROADCAST_B} == "true":
  XNN_TEST_BINARY_INPLACE_A(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
$else:
  XNN_TEST_BINARY_INPLACE_A(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
  XNN_TEST_BINARY_INPLACE_B(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
  XNN_TEST_BINARY_INPLACE_A_AND_B(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});

$if DATATYPE.startswith("q"):
  XNN_TEST_BINARY_A_ZERO_POINT(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
  XNN_TEST_BINARY_B_ZERO_POINT(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
  XNN_TEST_BINARY_Y_ZERO_POINT(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
  XNN_TEST_BINARY_A_SCALE(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
  XNN_TEST_BINARY_B_SCALE(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
  XNN_TEST_BINARY_Y_SCALE(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});

$if "minmax" in ACTIVATION_TYPE:
  XNN_TEST_BINARY_QMIN(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
  XNN_TEST_BINARY_QMAX(ukernel, arch_flags, batch_tile, ${BROADCAST_B}, datatype, ${", ".join(TEST_ARGS)});
"""

def main(args):
  options = parser.parse_args(args)

  tester = options.tester
  tester_header = {
    "VCMulMicrokernelTester": "vcmul-microkernel-tester.h",
    "VBinaryMicrokernelTester": "vbinary-microkernel-tester.h",
  }[tester]
  tests = """\
// Copyright 2019 Google LLC
//
// This source code is licensed under the BSD-style license found in the
// LICENSE file in the root directory of this source tree.
//
// Auto-generated file. Do not edit!
//   Microkernel: {microkernel}
//   Generator: {generator}


#include "xnnpack/microparams-init.h"
#include "xnnpack/vbinary.h"
#include "{tester_header}"

""".format(
      microkernel=options.ukernel,
      generator=sys.argv[0],
      tester_header=tester_header,
  )

  ukernel_parts = options.ukernel.split("-")
  datatype = ukernel_parts[0]
  op = ukernel_parts[1]
  activation = ukernel_parts[2] if len(ukernel_parts) >= 3 else ""

  broadcast_b = False
  if op[-1] == 'c':
    broadcast_b = True
  op_type = OP_TYPES[op]

  test_args = ["ukernel"]
  if tester in ["VBinaryMicrokernelTester"] and not datatype in ['qs8', 'qu8']:
    test_args.append("%s::OpType::%s" % (tester, op_type))
  test_args.append("init_params")
  tests += xnncommon.make_multiline_macro(xngen.preprocess(
      BINOP_TEST_TEMPLATE,
      {
          "TEST_ARGS": test_args,
          "TESTER": tester,
          "BROADCAST_B": str(broadcast_b).lower(),
          "DATATYPE": datatype,
          "OP_TYPE": op_type,
          "ACTIVATION_TYPE": activation,
      },
  ))

  folder = datatype + "-" + ("vbinary" if datatype.startswith("f") else op)
  tests += f'#include "{xnncommon.xnnpack_src()}{folder}/{options.ukernel}.h"\n'
  tests += "#undef XNN_UKERNEL_WITH_PARAMS\n"
  tests = tests.replace("s32-vmulc/s32-vmulc.h", "s32-vmul/s32-vmulc.h")

  xnncommon.overwrite_if_changed(options.output, tests)


if __name__ == "__main__":
  main(sys.argv[1:])



#!/usr/bin/env python
# Copyright 2019 Google LLC
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import os
import sys

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
import xngen
import xnncommon


parser = argparse.ArgumentParser(
    description="Vector unary operation microkernel test generator"
)
parser.add_argument(
    "-k",
    "--ukernel",
    required=True,
    help="Microkernel to generate",
)
parser.add_argument(
    "-o",
    "--output",
    metavar="FILE",
    required=True,
    help="Output (C++ source) file",
)
parser.set_defaults(defines=list())

OP_TYPES = {
    "vabs": "Abs",
    "vclamp": "Clamp",
    "velu": "ELU",
    "vexp": "Exp",
    "vgelu": "GELU",
    "vhswish": "HardSwish",
    "vlog": "Log",
    "vlrelu": "LeakyReLU",
    "vneg": "Negate",
    "vrelu": "ReLU",
    "vrndd": "RoundDown",
    "vrndne": "RoundToNearestEven",
    "vrndz": "RoundTowardsZero",
    "vrndu": "RoundUp",
    "vrsqrt": "ReciprocalSquareRoot",
    "vsigmoid": "Sigmoid",
    "vsqr": "Square",
    "vsqrt": "SquareRoot",
    "vtanh": "TanH",
}

PARAMS_TYPES = ["Clamp", "ELU", "LeakyReLU"]

SPECIAL_VALUES_F32 = {
    "SquareRoot": (
        4,  # Number of elements.
        "{0.0f, -0.0f, 1.0f, -1.0f}",  # Inputs.
        "{0.0f, -0.0f, 1.0f, NAN}",  # Expected outputs.
        1,  # Error margin in ULP.
    ),
    "TanH": (
        7,  # Number of elements.
        "{0.0f, -0.0f, 10.0f, -10.0f, INFINITY, -INFINITY, NAN}",
        "{0.0f, -0.0f, 1.0f, -1.0f, 1.0f, -1.0f, NAN}",
        # TODO: b/338934971 - This should be `1` ulp, but this fails on
        # `cmake-linux-riscv64-rvv` (but not on `cmake-linux-riscv64`).
        3,
    ),
    "Log": (
        4,  # Number of elements.
        "{1.0f, -1.0f, 0.0f, -0.0f}",  # Inputs.
        "{0.0f, NAN, -INFINITY, -INFINITY}",  # Expected outputs.
        1,  # Error margin in ULP.
    ),
    "GELU": (
        3,  # Number of elements.
        "{-6.0f, 6.0f, 0.0f}",  # Inputs.
        "{0.0f, 6.0f, 0.0f}",  # Expected outputs.
        1,  # Error margin in ULP.
    ),
    "Exp": (
        3,  # Number of elements.
        "{0.0f, -1e3f, 1e3f}",  # Inputs.
        "{1.0f, 0.0f, INFINITY}",  # Expected outputs.
        1,  # Error margin in ULP.
    ),
}

TEST_TEMPLATE = """\
#define XNN_UKERNEL_WITH_PARAMS(arch_flags, ukernel, batch_tile, vector_tile, datatype, params_type, init_params)
  TEST(ukernel, batch_eq) { TestBatchEq<TestInfo, datatype, datatype>(arch_flags, batch_tile, ukernel, init_params); }
  TEST(ukernel, batch_div) { TestBatchDiv<TestInfo, datatype, datatype>(arch_flags, batch_tile, ukernel, init_params); }
  TEST(ukernel, batch_lt) { TestBatchLT<TestInfo, datatype, datatype>(arch_flags, batch_tile, ukernel, init_params); }
  TEST(ukernel, batch_gt) { TestBatchGT<TestInfo, datatype, datatype>(arch_flags, batch_tile, ukernel, init_params); }
  TEST(ukernel, inplace) { TestInPlace<TestInfo, datatype, datatype>(arch_flags, batch_tile, ukernel, init_params); }
$if OP_TYPE == "Clamp":
  TEST(ukernel, clamp_min) {
    TEST_REQUIRES_ARCH_FLAGS(arch_flags);
    const size_t batch_scale = get_batch_scale<datatype>();
    const size_t batch_size = batch_tile * batch_scale;
    for (int16_t min : {-128, -20, -1, 0, 1, 30, 127, 255}) {
      xnn_unary_params params;
      params.clamp.min = min;
      params.clamp.max = 255;
      VUnaryMicrokernelTester()
          .batch_size(batch_size)
          .Test<TestInfo>(ukernel, init_params, params);
    }
  }

  TEST(ukernel, clamp_max) {
    TEST_REQUIRES_ARCH_FLAGS(arch_flags);
    const size_t batch_scale = get_batch_scale<datatype>();
    const size_t batch_size = batch_tile * batch_scale;
    for (int16_t max : {-127, -11, 0, 40, 127, 255}) {
      xnn_unary_params params;
      params.clamp.min = -128;
      params.clamp.max = max;
      VUnaryMicrokernelTester()
          .batch_size(batch_size)
          .Test<TestInfo>(ukernel, init_params, params);
    }
  }
$if OP_TYPE == "ELU":
  TEST(ukernel, alpha) {
    TEST_REQUIRES_ARCH_FLAGS(arch_flags);
    const size_t batch_scale = get_batch_scale<datatype>();
    const size_t batch_size = batch_tile * batch_scale;
    for (float alpha : std::array<float, 2>({0.3f, 3.0f})) {
      xnn_unary_params params;
      params.elu.alpha = alpha;
      ${TESTER}()
        .batch_size(batch_size)
        .Test<TestInfo>(ukernel, init_params, params);
    }
  }
$if OP_TYPE == "LeakyReLU":
  TEST(ukernel, negative_slope) {
    TEST_REQUIRES_ARCH_FLAGS(arch_flags);
    const size_t batch_scale = get_batch_scale<datatype>();
    const size_t batch_size = batch_tile * batch_scale;
    for (float negative_slope : {0.01f, 0.3f, 1.3f}) {
      xnn_unary_params params;
      params.leaky_relu.negative_slope = negative_slope;
      ${TESTER}()
        .batch_size(batch_size)
        .Test<TestInfo>(ukernel, init_params, params);
    }
  }
$if "q" in DATATYPE:
  TEST(ukernel, input_scale) { TestInputScale<TestInfo, datatype, datatype>(arch_flags, batch_tile, ukernel, init_params); }
  TEST(ukernel, output_scale) { TestOutputScale<TestInfo, datatype, datatype>(arch_flags, batch_tile, ukernel, init_params); }
  TEST(ukernel, input_zero_point) { TestInputZeroPoint<TestInfo, datatype, datatype>(arch_flags, batch_tile, ukernel, init_params); }
  TEST(ukernel, output_zero_point) { TestOutputZeroPoint<TestInfo, datatype, datatype>(arch_flags, batch_tile, ukernel, init_params); }
$if DATATYPE == "f32" and OP_TYPE in SPECIAL_VALUES_F32:
  TEST(ukernel, special_values) {
    TEST_REQUIRES_ARCH_FLAGS(arch_flags);
    VUnaryMicrokernelTester().Test<TestInfo>(ukernel, init_params,
      /*inputs=*/${SPECIAL_VALUES_F32[OP_TYPE][1]},
      /*outputs=*/${SPECIAL_VALUES_F32[OP_TYPE][2]},
      /*tolerance_ulp=*/${SPECIAL_VALUES_F32[OP_TYPE][3]});
  }
"""

def main(args):
  options = parser.parse_args(args)

  parts = options.ukernel.split("-")
  datatype = parts[-2]
  op = parts[-1]
  op_type = OP_TYPES[op]

  tester = "VUnaryMicrokernelTester"
  tester_header = "vunary-microkernel-tester.h"
  op_header = "vunary.h"
  tests = """\
// Copyright 2019 Google LLC
//
// This source code is licensed under the BSD-style license found in the
// LICENSE file in the root directory of this source tree.
//
// Auto-generated file. Do not edit!
//   Microkernel: {microkernel}
//   Generator: {generator}


#include <array>
#include <cmath>
#include <cstdint>
#include <cstddef>
#include <limits>

#include <gtest/gtest.h>
#include "xnnpack.h"
#include "xnnpack/common.h"
#include "xnnpack/isa-checks.h"
#include "xnnpack/microparams-init.h"
#include "xnnpack/microparams.h"
#include "xnnpack/{op_header}"
#include "next_prime.h"
#include "{tester_header}"

""".format(
      microkernel=options.ukernel,
      generator=sys.argv[0],
      op_header=op_header,
      tester_header=tester_header,
  )

  test_args = ["ukernel", "init_params"]

  tests += """\
using TestInfo = {op_type};

""".format(op_type=op_type)

  tests += xnncommon.make_multiline_macro(xngen.preprocess(
      TEST_TEMPLATE,
      {
          "TESTER": tester,
          "TEST_ARGS": test_args,
          "DATATYPE": datatype,
          "OP_TYPE": op_type,
          "OP_NAME": op,
          "SPECIAL_VALUES_F32": SPECIAL_VALUES_F32,
      },
  ))

  folder = options.ukernel
  if "rnd" in folder:
    folder = folder[0:8]

  tests += f'#include "{xnncommon.xnnpack_src()}{folder}/{options.ukernel}.h"\n'
  tests += "#undef XNN_UKERNEL_WITH_PARAMS\n"

  xnncommon.overwrite_if_changed(options.output, tests)


if __name__ == "__main__":
  main(sys.argv[1:])



#!/usr/bin/env python
# Copyright 2019 Google LLC
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import codecs
import os
import re
import sys
import yaml

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
from primes import next_prime
import xngen
import xnncommon


parser = argparse.ArgumentParser(
  description='IBILINEAR microkernel test generator')
parser.add_argument("-s", "--spec", metavar="FILE", required=True,
                    help="Specification (YAML) file")
parser.add_argument("-o", "--output", metavar="FILE", required=True,
                    help='Output (C++ source) file')
parser.set_defaults(defines=list())


def split_ukernel_name(name):
  match = re.fullmatch(r"xnn_(f16|f32|s8|u8)_ibilinear_ukernel__(.+)_c(\d+)", name)
  assert match is not None
  channel_tile = int(match.group(3))
  pixel_tile = 1

  arch, isa, assembly = xnncommon.parse_target_name(target_name=match.group(2))
  return channel_tile, pixel_tile, arch, isa


IBILINEAR_TEST_TEMPLATE = """\
TEST(${TEST_NAME}, channels_eq_${CHANNEL_TILE}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  IBilinearMicrokernelTester()
    .pixels(${PIXEL_TILE})
    .channels(${CHANNEL_TILE})
    .Test(${TEST_FUNC});
}

$if CHANNEL_TILE > 1:
  TEST(${TEST_NAME}, channels_div_${CHANNEL_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t channels = ${CHANNEL_TILE*2}; channels < ${CHANNEL_TILE*10}; channels += ${CHANNEL_TILE}) {
      IBilinearMicrokernelTester()
        .pixels(${PIXEL_TILE})
        .channels(channels)
        .Test(${TEST_FUNC});
    }
  }

  TEST(${TEST_NAME}, channels_lt_${CHANNEL_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t channels = 1; channels < ${CHANNEL_TILE}; channels++) {
      IBilinearMicrokernelTester()
        .pixels(${PIXEL_TILE})
        .channels(channels)
        .Test(${TEST_FUNC});
    }
  }

TEST(${TEST_NAME}, channels_gt_${CHANNEL_TILE}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t channels = ${CHANNEL_TILE+1}; channels < ${10 if CHANNEL_TILE == 1 else CHANNEL_TILE*2}; channels++) {
    IBilinearMicrokernelTester()
      .pixels(${PIXEL_TILE})
      .channels(channels)
      .Test(${TEST_FUNC});
  }
}

$if PIXEL_TILE > 1:
  TEST(${TEST_NAME}, pixels_div_${PIXEL_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t pixels = ${PIXEL_TILE*2}; pixels < ${PIXEL_TILE*10}; pixels += ${PIXEL_TILE}) {
      for (size_t channels = 1; channels <= ${CHANNEL_TILE * 5}; channels += ${max(1, CHANNEL_TILE - 1)}) {
        IBilinearMicrokernelTester()
          .pixels(pixels)
          .channels(channels)
          .Test(${TEST_FUNC});
      }
    }
  }

  TEST(${TEST_NAME}, pixels_lt_${PIXEL_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t pixels = 1; pixels < ${PIXEL_TILE}; pixels++) {
      for (size_t channels = 1; channels <= ${CHANNEL_TILE * 5}; channels += ${max(1, CHANNEL_TILE - 1)}) {
        IBilinearMicrokernelTester()
          .pixels(pixels)
          .channels(channels)
          .Test(${TEST_FUNC});
      }
    }
  }

TEST(${TEST_NAME}, pixels_gt_${PIXEL_TILE}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t pixels = ${PIXEL_TILE+1}; pixels < ${max(PIXEL_TILE*2, 3)}; pixels++) {
    for (size_t channels = 1; channels <= ${CHANNEL_TILE * 5}; channels += ${max(1, CHANNEL_TILE - 1)}) {
      IBilinearMicrokernelTester()
        .pixels(pixels)
        .channels(channels)
        .Test(${TEST_FUNC});
    }
  }
}

TEST(${TEST_NAME}, input_offset) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t pixels = 1; pixels < ${PIXEL_TILE * 5}; pixels += ${max(1, PIXEL_TILE - 1)}) {
    for (size_t channels = 1; channels <= ${CHANNEL_TILE * 5}; channels += ${max(1, CHANNEL_TILE - 1)}) {
      IBilinearMicrokernelTester()
        .pixels(pixels)
        .channels(channels)
        .input_offset(${next_prime(CHANNEL_TILE * 5 + 1)})
        .Test(${TEST_FUNC});
    }
  }
}

TEST(${TEST_NAME}, output_stride) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t pixels = 1; pixels < ${PIXEL_TILE * 5}; pixels += ${max(1, PIXEL_TILE - 1)}) {
    for (size_t channels = 1; channels <= ${CHANNEL_TILE * 5}; channels += ${max(1, CHANNEL_TILE - 1)}) {
      IBilinearMicrokernelTester()
        .pixels(pixels)
        .channels(channels)
        .output_stride(${next_prime(CHANNEL_TILE * 5 + 1)})
        .Test(${TEST_FUNC});
    }
  }
}
"""


def generate_test_cases(ukernel, channel_tile, pixel_tile, isa):
  """Generates all tests cases for a BILINEAR micro-kernel.

  Args:
    ukernel: C name of the micro-kernel function.
    channel_tile: Number of channels processed per one iteration of the inner
                  loop of the micro-kernel.
    pixel_tile: Number of pixels processed per one iteration of the outer loop
                of the micro-kernel.
    isa: instruction set required to run the micro-kernel. Generated unit test
         will skip execution if the host processor doesn't support this ISA.

  Returns:
    Code for the test case.
  """
  _, test_name = ukernel.split("_", 1)
  _, datatype, ukernel_type, _ = ukernel.split("_", 3)
  test_args = [ukernel]
  return xngen.preprocess(IBILINEAR_TEST_TEMPLATE, {
      "TEST_NAME": test_name.upper().replace("UKERNEL_", ""),
      "TEST_FUNC": ukernel,
      "UKERNEL_TYPE": ukernel_type.upper(),
      "DATATYPE": datatype,
      "CHANNEL_TILE": channel_tile,
      "PIXEL_TILE": pixel_tile,
      "ISA_CHECK": xnncommon.generate_isa_check_macro(isa),
      "next_prime": next_prime,
    })


def main(args):
  options = parser.parse_args(args)

  with codecs.open(options.spec, "r", encoding="utf-8") as spec_file:
    spec_yaml = yaml.safe_load(spec_file)
    if not isinstance(spec_yaml, list):
      raise ValueError("expected a list of micro-kernels in the spec")

    tests = """\
// Copyright 2019 Google LLC
//
// This source code is licensed under the BSD-style license found in the
// LICENSE file in the root directory of this source tree.
//
// Auto-generated file. Do not edit!
//   Specification: {specification}
//   Generator: {generator}


#include <gtest/gtest.h>
#include "xnnpack/common.h"
#include "xnnpack/ibilinear.h"
#include "xnnpack/isa-checks.h"
#include "ibilinear-microkernel-tester.h"
""".format(specification=options.spec, generator=sys.argv[0])

    for ukernel_spec in spec_yaml:
      name = ukernel_spec["name"]
      channel_tile, pixel_tile, arch, isa = split_ukernel_name(name)

      test_case = generate_test_cases(name, channel_tile, pixel_tile, isa)
      tests += "\n\n" + xnncommon.postprocess_test_case(test_case, arch, isa)

    xnncommon.overwrite_if_changed(options.output, tests)


if __name__ == "__main__":
  main(sys.argv[1:])



#!/usr/bin/python3

import os
import argparse

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('filename', help='Scripts to read', type=str)
    parser.add_argument('--files-contain', dest='files_contain', default='/', type=str)
    args = parser.parse_args()

    script_to_check = args.filename
    file_paths_contain = args.files_contain

    file_edit_dates = {}
    outputs = {}
    files_exist = {}
    with open(script_to_check, 'r') as f:
        for line in f.readlines():
            if line.lstrip().startswith('#'):continue
            file_names = [x.strip() for x in line.split(' ')]
            # Filter file names based on presence of 'files_contain'
            file_names = [x for x in file_names if file_paths_contain in x]
            if len(file_names) >= 3:
                # outputs[file_names[1]] = file_names[0]
                # i==0 is script. i==1 is input file. i>=1 is for output files
                for i, file_name in enumerate(file_names):
                    if file_name not in files_exist:
                        files_exist[file_name] = os.path.exists(file_name)

                    # if file doesn't exist
                    if not files_exist[file_name]:
                        # if output file does not exist
                        if i > 1:
                            file_edit_dates[file_name] = 0
                        else:
                            continue
                    if file_name not in file_edit_dates:
                        file_edit_dates[file_name] = os.path.getmtime(file_name)
                    if i > 1:
                        outputs[file_name] = file_names[1]

    for out_file, in_file in outputs.items():
        if file_edit_dates[in_file] > file_edit_dates[out_file]:
            print(out_file)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright 2021 Google LLC
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import codecs
import math
import os
import re
import sys
import yaml

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
import xngen
import xnncommon


parser = argparse.ArgumentParser(
  description='LUT microkernel test generator')
parser.add_argument("-s", "--spec", metavar="FILE", required=True,
                    help="Specification (YAML) file")
parser.add_argument("-o", "--output", metavar="FILE", required=True,
                    help='Output (C++ source) file')
parser.set_defaults(defines=list())


def split_ukernel_name(name):
  match = re.fullmatch(r"xnn_x8_lut_ukernel__(.+)_u(\d+)(v)?", name)
  if match is None:
    raise ValueError("Unexpected microkernel name: " + name)
  batch_tile = int(match.group(2))

  arch, isa, assembly = xnncommon.parse_target_name(target_name=match.group(1))
  return batch_tile, arch, isa


LUT_TEST_TEMPLATE = """\
TEST(${TEST_NAME}, batch_eq_${BATCH_TILE}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  LUTMicrokernelTester()
    .batch_size(${BATCH_TILE})
    .Test(${UKERNEL_NAME});
}

$if BATCH_TILE > 1:
  TEST(${TEST_NAME}, batch_div_${BATCH_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t batch_size = ${BATCH_TILE*2}; batch_size < ${BATCH_TILE*10}; batch_size += ${BATCH_TILE}) {
      LUTMicrokernelTester()
        .batch_size(batch_size)
        .Test(${UKERNEL_NAME});
    }
  }

  TEST(${TEST_NAME}, batch_lt_${BATCH_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t batch_size = 1; batch_size < ${BATCH_TILE}; batch_size++) {
      LUTMicrokernelTester()
        .batch_size(batch_size)
        .Test(${UKERNEL_NAME});
    }
  }

TEST(${TEST_NAME}, batch_gt_${BATCH_TILE}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t batch_size = ${BATCH_TILE+1}; batch_size < ${10 if BATCH_TILE == 1 else BATCH_TILE*2}; batch_size++) {
    LUTMicrokernelTester()
      .batch_size(batch_size)
      .Test(${UKERNEL_NAME});
  }
}

TEST(${TEST_NAME}, inplace) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t batch_size = 1; batch_size <= ${BATCH_TILE*5}; batch_size += ${max(1, BATCH_TILE-1)}) {
    LUTMicrokernelTester()
      .batch_size(batch_size)
      .inplace(true)
      .Test(${UKERNEL_NAME});
  }
}
"""


def generate_test_cases(ukernel, batch_tile, isa):
  """Generates all tests cases for a LUT micro-kernel.

  Args:
    ukernel: C name of the micro-kernel function.
    batch_tile: Number of batch elements processed per one iteration of the
                inner loop of the micro-kernel.
    isa: instruction set required to run the micro-kernel. Generated unit test
         will skip execution if the host processor doesn't support this ISA.

  Returns:
    Code for the test case.
  """
  _, test_name = ukernel.split("_", 1)
  return xngen.preprocess(LUT_TEST_TEMPLATE, {
      "TEST_NAME": test_name.upper().replace("UKERNEL_", ""),
      "BATCH_TILE": batch_tile,
      "UKERNEL_NAME": ukernel,
      "ISA_CHECK": xnncommon.generate_isa_check_macro(isa),
    })


def main(args):
  options = parser.parse_args(args)

  with codecs.open(options.spec, "r", encoding="utf-8") as spec_file:
    spec_yaml = yaml.safe_load(spec_file)
    if not isinstance(spec_yaml, list):
      raise ValueError("expected a list of micro-kernels in the spec")

    tests = """\
// Copyright 2021 Google LLC
//
// This source code is licensed under the BSD-style license found in the
// LICENSE file in the root directory of this source tree.
//
// Auto-generated file. Do not edit!
//   Specification: {specification}
//   Generator: {generator}


#include <gtest/gtest.h>
#include "xnnpack/common.h"
#include "xnnpack/isa-checks.h"
#include "xnnpack/lut.h"
#include "lut-microkernel-tester.h"
""".format(specification=options.spec, generator=sys.argv[0])

    for ukernel_spec in spec_yaml:
      name = ukernel_spec["name"]
      batch_tile, arch, isa = split_ukernel_name(name)

      test_case = generate_test_cases(name, batch_tile, isa)
      tests += "\n\n" + xnncommon.postprocess_test_case(test_case, arch, isa)

    xnncommon.overwrite_if_changed(options.output, tests)


if __name__ == "__main__":
  main(sys.argv[1:])



#!/usr/bin/env python
# Copyright 2020 Google LLC
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import codecs
import os
import re
import sys
import yaml

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
from primes import next_prime
import xngen
import xnncommon


parser = argparse.ArgumentParser(
  description='Test generator for DWCONV2D CHW micro-kernels')
parser.add_argument("-s", "--spec", metavar="FILE", required=True,
                    help="Spec (YAML) file")
parser.add_argument("-o", "--output", metavar="FILE", required=True,
                    help='Output (C++ source) file')
parser.set_defaults(defines=list())


TEST_TEMPLATE = """\
$if SUBSAMPLING == 1:
  TEST(${TEST_NAME}, output_width_eq_${WIDTH_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    DWConv2DMicrokernelTester()
      .input_width(${(WIDTH_TILE - 1) * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING})
      .input_height(${HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING - 1})
      .kernel_height(${KERNEL_HEIGHT})
      .kernel_width(${KERNEL_WIDTH})
      .subsampling(${SUBSAMPLING})
      .padding_left(${PADDING})
      .padding_right(${PADDING})
      .padding_top(${PADDING})
      .padding_bottom(${PADDING})
      .Test(${", ".join(TEST_ARGS)});
  }
$else:
  TEST(${TEST_NAME}, output_width_eq_${WIDTH_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t input_width = ${(WIDTH_TILE - 1) * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING}; input_width < ${WIDTH_TILE * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING}; input_width++) {
      DWConv2DMicrokernelTester()
        .input_width(input_width)
        .input_height(${HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING - 1})
        .kernel_height(${KERNEL_HEIGHT})
        .kernel_width(${KERNEL_WIDTH})
        .subsampling(${SUBSAMPLING})
        .padding_left(${PADDING})
        .padding_right(${PADDING})
        .padding_top(${PADDING})
        .padding_bottom(${PADDING})
        .Test(${", ".join(TEST_ARGS)});
    }
  }

$if WIDTH_TILE > 1:
  TEST(${TEST_NAME}, output_width_div_${WIDTH_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t input_width = ${2 * WIDTH_TILE * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING - 1}; input_width < ${8 * WIDTH_TILE * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING - 1}; input_width += ${WIDTH_TILE * SUBSAMPLING}) {
      DWConv2DMicrokernelTester()
        .input_width(input_width)
        .input_height(${HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING - 1})
        .kernel_height(${KERNEL_HEIGHT})
        .kernel_width(${KERNEL_WIDTH})
        .subsampling(${SUBSAMPLING})
        .padding_left(${PADDING})
        .padding_right(${PADDING})
        .padding_top(${PADDING})
        .padding_bottom(${PADDING})
        .Test(${", ".join(TEST_ARGS)});
    }
  }

  TEST(${TEST_NAME}, output_width_lt_${WIDTH_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t input_width = ${max(1, KERNEL_WIDTH - 2 * PADDING)}; input_width < ${(WIDTH_TILE - 1) * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING}; input_width++) {
      DWConv2DMicrokernelTester()
        .input_width(${WIDTH_TILE * SUBSAMPLING})
        .input_height(${HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING - 1})
        .kernel_height(${KERNEL_HEIGHT})
        .kernel_width(${KERNEL_WIDTH})
        .subsampling(${SUBSAMPLING})
        .padding_left(${PADDING})
        .padding_right(${PADDING})
        .padding_top(${PADDING})
        .padding_bottom(${PADDING})
        .Test(${", ".join(TEST_ARGS)});
    }
  }

TEST(${TEST_NAME}, output_width_gt_${WIDTH_TILE}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t input_width = ${WIDTH_TILE * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING}; input_width < ${(5 if WIDTH_TILE == 1 else 2) * WIDTH_TILE * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING}; input_width++) {
    DWConv2DMicrokernelTester()
      .input_width(input_width)
      .input_height(${HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING - 1})
      .kernel_height(${KERNEL_HEIGHT})
      .kernel_width(${KERNEL_WIDTH})
      .subsampling(${SUBSAMPLING})
      .padding_left(${PADDING})
      .padding_right(${PADDING})
      .padding_top(${PADDING})
      .padding_bottom(${PADDING})
      .Test(${", ".join(TEST_ARGS)});
  }
}

$if SUBSAMPLING > 1:
  TEST(${TEST_NAME}, output_height_eq_${HEIGHT_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t input_height = ${(HEIGHT_TILE - 1) * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING}; input_height < ${HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING}; input_height++) {
      for (size_t input_width = 1; input_width < ${5 * WIDTH_TILE * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING}; input_width += ${max(1, WIDTH_TILE * SUBSAMPLING - 1)}) {
        DWConv2DMicrokernelTester()
          .input_width(input_width)
          .input_height(input_height)
          .kernel_height(${KERNEL_HEIGHT})
          .kernel_width(${KERNEL_WIDTH})
          .subsampling(${SUBSAMPLING})
          .padding_left(${PADDING})
          .padding_right(${PADDING})
          .padding_top(${PADDING})
          .padding_bottom(${PADDING})
          .Test(${", ".join(TEST_ARGS)});
      }
    }
  }

$if HEIGHT_TILE > 1:
  TEST(${TEST_NAME}, output_height_div_${HEIGHT_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t input_height = ${2 * HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING - 1}; input_height < ${8 * HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING - 1}; input_height += ${HEIGHT_TILE * SUBSAMPLING}) {
      for (size_t input_width = 1; input_width < ${5 * WIDTH_TILE * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING}; input_width += ${max(1, WIDTH_TILE * SUBSAMPLING - 1)}) {
        DWConv2DMicrokernelTester()
          .input_width(input_width)
          .input_height(input_height)
          .kernel_height(${KERNEL_HEIGHT})
          .kernel_width(${KERNEL_WIDTH})
          .subsampling(${SUBSAMPLING})
          .padding_left(${PADDING})
          .padding_right(${PADDING})
          .padding_top(${PADDING})
          .padding_bottom(${PADDING})
          .Test(${", ".join(TEST_ARGS)});
      }
    }
  }

  TEST(${TEST_NAME}, output_height_lt_${HEIGHT_TILE}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t input_height = ${max(1, KERNEL_HEIGHT - 2 * PADDING)}; input_height < ${(HEIGHT_TILE - 1) * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING}; input_height++) {
      for (size_t input_width = 1; input_width < ${5 * WIDTH_TILE * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING}; input_width += ${max(1, WIDTH_TILE * SUBSAMPLING - 1)}) {
        DWConv2DMicrokernelTester()
          .input_width(input_width)
          .input_height(input_height)
          .kernel_height(${KERNEL_HEIGHT})
          .kernel_width(${KERNEL_WIDTH})
          .subsampling(${SUBSAMPLING})
          .padding_left(${PADDING})
          .padding_right(${PADDING})
          .padding_top(${PADDING})
          .padding_bottom(${PADDING})
          .Test(${", ".join(TEST_ARGS)});
      }
    }
  }

TEST(${TEST_NAME}, output_height_gt_${HEIGHT_TILE}) {
  $if ISA_CHECK:
    ${ISA_CHECK};
  for (size_t input_height = ${HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING}; input_height < ${(5 if WIDTH_TILE == 1 else 2) * HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING}; input_height++) {
    for (size_t input_width = 1; input_width < ${5 * WIDTH_TILE * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING}; input_width += ${max(1, WIDTH_TILE * SUBSAMPLING - 1)}) {
      DWConv2DMicrokernelTester()
        .input_width(input_width)
        .input_height(input_height)
        .kernel_height(${KERNEL_HEIGHT})
        .kernel_width(${KERNEL_WIDTH})
        .subsampling(${SUBSAMPLING})
        .padding_left(${PADDING})
        .padding_right(${PADDING})
        .padding_top(${PADDING})
        .padding_bottom(${PADDING})
        .Test(${", ".join(TEST_ARGS)});
    }
  }
}

$if SUBSAMPLING > 1:
  TEST(${TEST_NAME}, padding_top_eq_${SUBSAMPLING - 1}) {
    $if ISA_CHECK:
      ${ISA_CHECK};
    for (size_t input_height = ${max(1, KERNEL_HEIGHT - 2 * PADDING + 1)}; input_height < ${3 * HEIGHT_TILE * SUBSAMPLING + KERNEL_HEIGHT - 2 * PADDING + 1}; input_height++) {
      for (size_t input_width = 1; input_width < ${5 * WIDTH_TILE * SUBSAMPLING + KERNEL_WIDTH - 2 * PADDING}; input_width += ${max(1, WIDTH_TILE * SUBSAMPLING - 1)}) {
        DWConv2DMicrokernelTester()
          .input_width(input_width)
          .input_height(input_height)
          .kernel_height(${KERNEL_HEIGHT})
          .kernel_width(${KERNEL_WIDTH})
          .subsampling(${SUBSAMPLING})
          .padding_left(${PADDING})
          .padding_right(${PADDING})
          .padding_top(${PADDING - 1})
          .padding_bottom(${PADDING})
          .Test(${", ".join(TEST_ARGS)});
      }
    }
  }
"""

def split_ukernel_name(name):
  match = re.fullmatch(r"xnn_(f16|f32)_dwconv2d_chw_ukernel_(\d+)x(\d+)(s2)?p(\d+)__(.+)_(\d+)x(\d+)(_acc\d+)?", name)
  assert match is not None
  kernel_height, kernel_width = int(match.group(2)), int(match.group(3))
  if match.group(4):
    assert match.group(4).startswith("s")
    stride = int(match.group(4)[1:])
  else:
    stride = 1
  padding = int(match.group(5))

  height_tile = int(match.group(7))
  width_tile = int(match.group(8))

  arch, isa, assembly = xnncommon.parse_target_name(target_name=match.group(6))
  return kernel_height, kernel_width, stride, padding, arch, isa, \
         height_tile, width_tile


def generate_test_cases(ukernel, kernel_height, kernel_width, subsampling, \
  init_fn, padding, isa, height_tile, width_tile):
  """Generates all tests cases for a DWCONV2D CHW micro-kernel.

  Args:
    ukernel: C name of the micro-kernel function.
    kernel_height: convolution kernel height assumed by the micro-kernel.
    kernel_width: convolution kernel width assumed by the micro-kernel.
    subsampling: convolution subsampling (stride) assumed by the micro-kernel.
                 The same subsampling factor is assumed for both horizontal and
                 vertical directions.
    init_fn: C name of the function to initialize microkernel parameters.
    padding: convolution padding value assumed by the micro-kernel for right,
             bottom, and left padding. If convolution stride is 1, the same
             padding value is assumed for the top padding. If convolution stride
             is different than 1, top padding is specified via micro-kernel
             parameter, and can be either padding or (padding - 1).
    isa: instruction set required to run the micro-kernel. Generated unit test
         will skip execution if the host processor doesn't support this ISA.
    height_tile: number of output rows processed in one iteration of the main
                 loop of the micro-kernel.
    width_tile: number of output columns processed in one iteration of the main
                loop of the micro-kernel.

  Returns:
    Code for the test case.
  """
  _, test_name = ukernel.split("_", 1)
  _, datatype, ukernel_type, _ = ukernel.split("_", 3)
  test_args = [ukernel, init_fn]
  return xngen.preprocess(TEST_TEMPLATE, {
      "TEST_NAME": test_name.upper().replace("UKERNEL_", ""),
      "TEST_ARGS": test_args,
      "UKERNEL_TYPE": ukernel_type.upper(),
      "DATATYPE": datatype,
      "KERNEL_HEIGHT": kernel_height,
      "KERNEL_WIDTH": kernel_width,
      "SUBSAMPLING": subsampling,
      "PADDING": padding,
      "HEIGHT_TILE": height_tile,
      "WIDTH_TILE": width_tile,
      "ISA_CHECK": xnncommon.generate_isa_check_macro(isa),
      "next_prime": next_prime,
    })


def main(args):
  options = parser.parse_args(args)

  with codecs.open(options.spec, "r", encoding="utf-8") as spec_file:
    spec_yaml = yaml.safe_load(spec_file)
    if not isinstance(spec_yaml, list):
      raise ValueError("expected a list of micro-kernels in the spec")

    tests = """\
// Copyright 2020 Google LLC
//
// This source code is licensed under the BSD-style license found in the
// LICENSE file in the root directory of this source tree.
//
// Auto-generated file. Do not edit!
//   Specification: {specification}
//   Generator: {generator}


#include <gtest/gtest.h>
#include "xnnpack/common.h"
#include "xnnpack/dwconv.h"
#include "xnnpack/isa-checks.h"
#include "xnnpack/microparams-init.h"
#include "dwconv2d-microkernel-tester.h"
""".format(specification=options.spec, generator=sys.argv[0])

    for ukernel_spec in spec_yaml:
      name = ukernel_spec["name"]
      init_fn = ukernel_spec["init"]
      pipelined = bool(ukernel_spec.get("pipelined", False))
      kernel_height, kernel_width, subsampling, padding, arch, isa, \
        height_tile, width_tile = split_ukernel_name(name)

      test_case = generate_test_cases(name, kernel_height, kernel_width, \
                                      subsampling, init_fn, padding, isa, \
                                      height_tile, width_tile)
      tests += "\n\n" + xnncommon.postprocess_test_case(test_case, arch, isa)

    xnncommon.overwrite_if_changed(options.output, tests)


if __name__ == "__main__":
  main(sys.argv[1:])



#!/usr/bin/python3
import re
import argparse
from functools import cmp_to_key

parser = argparse.ArgumentParser()
parser.add_argument('inputs', nargs='+', help='Input filename')
parser.add_argument('--min_lines', type=int, default=1, help='Specify the min number of lines to rearrange')
args = parser.parse_args()

min_num_lines = args.min_lines

filename_regex = r'^[^\(]*(src|source|unit_test|include)/.*$'

# caveat: Does fully not handle numbers with different number of digits.
def compare_strs(str1, str2, num_sort=True):
    winner = 0 # which one's bigger
    num_sort_mode = False
    for i in range(min(len(str1),len(str2))):
        if winner == 0:
            if num_sort:
                if str1[i].isdigit() and str2[i].isdigit():
                    if num_sort_mode:
                        str1_num += str1[i]
                        str2_num += str2[i]
                    else:
                        num_sort_mode = True
                        str1_num = str1[i]
                        str2_num = str2[i]
                elif num_sort_mode:
                    if str1[i].isdigit(): str1_num += str1[i]
                    if str2[i].isdigit(): str2_num += str2[i]
                    winner = int(str1_num) - int(str2_num)
                    num_sort_mode = False
                    if winner != 0: break
            if str1[i] != str2[i] and not num_sort_mode:
                if str2[i].isdigit():
                    winner = -1
                elif str1[i].isdigit() or str1[i] > str2[i]:
                    winner = 1
                else:
                    winner = -1
    if num_sort_mode: # if it's still in num_sort_mode then find the results
        if len(str1)-1 > i:
            i += 1
            if str1[i].isdigit(): str1_num += str1[i]
        elif len(str2)-1 > i:
            i += 1
            if str2[i].isdigit(): str2_num += str2[i]
        if int(str1_num) > int(str2_num):
            winner = 1
        elif int(str1_num) < int(str2_num):
            winner = -1
        num_sort_mode = False

    if winner == 0:
        if str2[i].isdigit():
            winner = -1
        elif str1[i].isdigit() or str1[i] > str2[i]:
            winner = 1
        else:
            winner = -1
    return winner


def custom_sort(listy):
    for _ in range(len(listy)-1+1):
        for i in range(len(listy)-1):
            bigger = compare_strs(listy[i], listy[i+1])
            if bigger > 0:
                tmp_str = listy[i]
                listy[i] = listy[i+1]
                listy[i+1] = tmp_str
    return listy

# file_lines = [l.strip(stripchars) for l in file_lines]
# currently sorting by first number in filename. smallest numbers first
# this thing may need the ability to recognize filenames
# file_lines = sorted(file_lines, key=lambda line: int(re.search('\d+', line.split('/')[-1] + ' 0').group(0)))
input_filenames=args.inputs
for input_filename in input_filenames:
    with open(input_filename,'r') as file:
        file_lines = file.read().split('\n')
    output_filename=input_filename
    files_range = [] # is a tuple
    filenames_ranges = []
    for i, line in enumerate(file_lines):
        if re.search(filename_regex, line):
            if files_range == []:
                files_range.append(i)
        else:
            if files_range != []:
                files_range.append(i-1)
                # manages min num lines
                if files_range[1] - files_range[0] + 1 >= min_num_lines:filenames_ranges.append(files_range)
                files_range = []
    if files_range != []:
        files_range.append(i-1)
        # manages min num lines
        if files_range[1] - files_range[0] + 1 >= min_num_lines:filenames_ranges.append(files_range)
        files_range = []


    for x in filenames_ranges:
        # accommodate for a close bracket at the end of the last filename
        if file_lines[x[1]][-1] == ')':
            append_chars = ')'
            file_lines[x[1]] = file_lines[x[1]][:-1]
        else:
            append_chars = ''

        new_lines = sorted(file_lines[x[0]:x[1]+1], key=cmp_to_key(compare_strs))
        for i, l in enumerate(range(x[0], x[1]+1)):
            file_lines[l] = new_lines[i]
            if l == x[1]:
                file_lines[l] += append_chars

    with open(output_filename, 'w') as file:
        file.write('\n'.join(file_lines))
